from logging import Logger
from threading import Lock
from typing import Dict

from sidecar.aws_session import AwsSession
from sidecar.cloud_logger import FileLogger
from sidecar.cloud_logger.file_logger import ICloudLogger
from sidecar.model.objects import ISidecarConfiguration
from .stream_writer import CloudWatchStreamWriter
from ..logs import AppLogEntry


class CloudWatchLogger(ICloudLogger):
    def __init__(self, config: ISidecarConfiguration, logger: Logger, aws_session: AwsSession, file_logger: FileLogger):
        cloud_external_key = config.cloud_external_key
        self.sandbox_id = config.sandbox_id
        self.logger = logger
        self.logs = aws_session.get_cloudwatch_client() if aws_session else None
        self.group_name = "colony-{cloud_external_key}".format(cloud_external_key=cloud_external_key)
        self._stream_writers = dict()  # type: Dict[str, CloudWatchStreamWriter]
        self._lock = Lock()
        self.file_logger = file_logger

    def _get_writer(self, log_entry: AppLogEntry) -> CloudWatchStreamWriter:
        stream_name = "/{sandbox_id}/{app_name}/{instance_id}/{log_type}".format(sandbox_id=self.sandbox_id,
                                                                                 app_name=log_entry.app,
                                                                                 instance_id=log_entry.instance,
                                                                                 log_type=log_entry.topic)
        writer = self._stream_writers.get(stream_name)
        if writer is None:
            with self._lock:
                writer = self._stream_writers.get(stream_name)
                if writer is None:
                    writer = CloudWatchStreamWriter(self.logs, self.group_name, stream_name, self.logger)
                self._stream_writers[stream_name] = writer
        return writer

    def write(self, log_entry: AppLogEntry):
        writer = self._get_writer(log_entry)
        writer.write(log_entry.log_events)
        self.file_logger.write(log_entry)



import json
from typing import Dict


class DeploymentOutputConverter:
    @staticmethod
    def convert_from_terraform_output(output_json: {}) -> Dict[str, str]:
        res = {}
        for k, v in output_json.items():
            output_type = v['type']
            if output_type == 'map' or output_type == 'list':
                res[k] = json.dumps(v['value'])
            else:
                res[k] = v['value']
        return res

    @staticmethod
    def convert_from_configuration_script(output: str) -> Dict[str, str]:
        res = dict()
        for line in output.splitlines():
            split = str.split(line, '=')
            name = split[0].strip()
            value = split[1].strip()
            res[name] = value
        return res



from abc import ABCMeta, abstractmethod
from typing import List


class DeploymentOutputStore(metaclass=ABCMeta):

    @abstractmethod
    def save_application_output(self, app_name: str, app_instance_id: str, output: str):
        raise NotImplemented()

    @abstractmethod
    def save_service_output(self, service_name: str, output_json: {}):
        raise NotImplemented()

    @abstractmethod
    def get_deployment_outputs(self, outputs: List[str]) -> str:
        raise NotImplemented()




import json
from typing import List

import jsonpickle
from azure.common import AzureMissingResourceHttpError
from azure.cosmosdb.table import Entity
from azure.cosmosdb.table.tableservice import TableService

from sidecar.azure_clp.data_store_service import DataStoreService


class AzureTableService(DataStoreService):

    def __init__(self, table_service: TableService,
                 default_table_name: str = "colonySandboxes",
                 default_partition_key: str = "colonySandbox"):
        super().__init__()
        self.default_partition_key = default_partition_key
        self.table_service = table_service
        self.default_table_name = default_table_name

    def create_table(self):
        # create the table if not exists
        if not self.table_service.exists(table_name=self.default_table_name):
            self.table_service.create_table(table_name=self.default_table_name,
                                            fail_on_exist=False)

    def insert_data(self, data: dict, data_id: str = None):
        entity = self._to_entity(data=data, data_id=data_id)
        self.table_service.insert_entity(table_name=self.default_table_name,
                                         entity=entity)

    def find_data_by_id(self, data_id: str) -> dict:
        entity = {}
        try:
            entity = self.table_service.get_entity(table_name=self.default_table_name,
                                                   row_key=data_id,
                                                   partition_key=self.default_partition_key)
        except AzureMissingResourceHttpError:
            return {}

        res = self._convert_entity_to_dict(entity)
        return res

    def find_data_by_ids(self, id_column_name: str, ids: List[str] = None) -> List[dict]:
        if not ids:
            return list()
        ids_filter = " or ".join([id_column_name.replace("-", "__") + " eq '{0}' ".format(str(iid)) for iid in ids])
        query_entities_gen = self.table_service.query_entities(table_name=self.default_table_name,
                                                               filter=ids_filter)

        return list([self._convert_entity_to_dict(entity) for entity in query_entities_gen])

    def delete_data_by_id(self, data_id: str) -> bool:
        entity = self.table_service.delete_entity(table_name=self.default_table_name,
                                                  partition_key=self.default_partition_key,
                                                  row_key=data_id)
        return entity is None

    def update_data(self, data_id: str, data: dict, column_name: str):
        entity = self._to_entity(data={column_name: data},
                                 data_id=data_id)

        self.table_service.merge_entity(table_name=self.default_table_name,
                                        entity=entity)

    def _to_entity(self, data: dict, data_id: str):
        task = Entity()
        task.PartitionKey = self.default_partition_key
        task.RowKey = str(data_id)
        for k, v in data.items():
            task[str(k).replace("-", "__")] = jsonpickle.encode(v)
        return task

    def _convert_entity_to_dict(self, entity):
        res = {}
        remove_keys = {'PartitionKey', "RowKey", "Timestamp", "etag"}
        for k, v in entity.items():
            key = str(str(k).replace("__", "-"))
            if key not in remove_keys:
                res[key] = jsonpickle.decode(v)
        return res



import json
import shlex
import threading
from logging import Logger
from typing import List

from sidecar.aws_session import AwsSession
from sidecar.aws_status_maintainer import AWSStatusMaintainer
from sidecar.aws_tag_helper import AwsTagHelper
from sidecar.const import Const
from sidecar.model.objects import ISidecarConfiguration
from sidecar.services.deployment_output_converter import DeploymentOutputConverter
from sidecar.services.deployment_output_store import DeploymentOutputStore


class AwsDeploymentOutputStore(DeploymentOutputStore):

    def __init__(self, logger: Logger,
                 aws_session: AwsSession,
                 status_maintainer: AWSStatusMaintainer,
                 config: ISidecarConfiguration):
        self._config = config
        self._aws_session = aws_session
        self._status_maintainer = status_maintainer
        self._logger = logger
        self._lock = threading.RLock()

    def _get_service_declared_outputs(self, service_name: str) -> List[str]:
        outputs = next((s.outputs for s in self._config.services if s.name == service_name))
        if outputs:
            return outputs
        return []

    def _get_application_declared_outputs(self, application_name: str) -> List[str]:
        outputs = next((a.outputs for a in self._config.apps if a.name == application_name))
        if outputs:
            return outputs
        return []

    @staticmethod
    def _parse_output_name(output: str):
        try:
            split = output.split('.')
            entity_name = split[1]
            output_name = split[2]
            return entity_name, output_name
        except IndexError:
            raise Exception(f"output {output} cannot be resolved")

    def get_deployment_outputs(self, outputs: List[str]) -> str:
        entity_name_outputs_map = {}
        res = {}
        for output in outputs:
            try:
                entity_name, output_name = self._parse_output_name(output)
                entity_outputs = entity_name_outputs_map.get(entity_name, None)
                if entity_outputs is None:
                    entity_outputs = self._status_maintainer.get_deployment_output(entity_name)
                    entity_name_outputs_map[entity_name] = entity_outputs

                res[output] = entity_outputs.get(output_name, 'Output value not found')
            except Exception as ex:
                res[output] = repr(ex)
        return "\n".join("{}={}".format(k, shlex.quote(v)) for k, v in res.items())

    @staticmethod
    def filter_redundant_deployment_outputs(deployment_output: dict, declared_outputs: List[str]):
        deployment_output_names = list(deployment_output.keys())
        for deployment_output_name in deployment_output_names:
            if deployment_output_name not in declared_outputs:
                deployment_output.pop(deployment_output_name, None)

    def save_service_output(self, service_name: str, output_json: {}):
        try:
            deployment_output = DeploymentOutputConverter.convert_from_terraform_output(output_json)
        except Exception:
            output_str = json.dumps(output_json)
            err = f"service '{service_name}' deployment output is not valid:\n{output_str}"
            self._logger.exception(err)
            raise Exception(err)

        self._log_deployment_output(service_name, deployment_output)
        declared_outputs = self._get_service_declared_outputs(service_name)
        self.filter_redundant_deployment_outputs(deployment_output, declared_outputs)
        with self._lock:
            self._status_maintainer.update_service_output(service_name, deployment_output)

    def _log_deployment_output(self, service_name: str, deployment_output: {}):
        output_str = json.dumps(deployment_output, indent=2)
        self._logger.info(f"service '{service_name}' deployment output is:\n{output_str}")

    def save_application_output(self, app_name: str, app_instance_id: str, output: str):
        try:
            deployment_output = DeploymentOutputConverter.convert_from_configuration_script(output)
        except Exception as exc:
            self._logger.exception(f"application '{app_instance_id}/{app_name}' "
                                   f"deployment output is not valid:\n{output}")
            raise exc

        self._logger.info(f"application '{app_instance_id}/{app_name}' deployment output is:\n{deployment_output}")
        declared_outputs = self._get_application_declared_outputs(app_name)
        self.filter_redundant_deployment_outputs(deployment_output, declared_outputs)

        instance = self._get_instance_by_id(instance_id=app_instance_id)
        instance_logical_id = AwsTagHelper.wait_for_tag(instance, Const.INSTANCELOGICALID, self._logger)
        with self._lock:
            self._status_maintainer.update_app_instance_output(instance_logical_id,
                                                               app_instance_id,
                                                               app_name,
                                                               deployment_output)

    def _get_instance_by_id(self, instance_id: str):
        ec2 = self._aws_session.get_ec2_resource()
        return ec2.Instance(instance_id)
